@@ -163,17 +163,19 @@ void crypto_init() {
 	}
 }
 
-/* returns DROPBEAR_SUCCESS if we have a match for algo, DROPBEAR_FAILURE
+/* algolen specifies the length of algo, algos is our local list to match
+ * against.
+ * Returns DROPBEAR_SUCCESS if we have a match for algo, DROPBEAR_FAILURE
  * otherwise */
 int have_algo(char* algo, size_t algolen, algo_type algos[]) {
 
-	int i = 0;
-	while (algos[i].name != NULL) {
+	int i;
+
+	for (i = 0; algos[i].name != NULL; i++) {
 		if (strlen(algos[i].name) == algolen
 				&& (strncmp(algos[i].name, algo, algolen) == 0)) {
 			return DROPBEAR_SUCCESS;
 		}
-		i++;
 	}
 
 	return DROPBEAR_FAILURE;
@@ -207,7 +209,7 @@ algo_type * buf_match_algo(buffer* buf, algo_type localalgos[]) {
 			/* someone is trying something strange */
 			goto out;
 		}
-		if (algolist[i] == ',' && i != len) {
+		if (algolist[i] == ',') {
 			algolist[i] = '\0';
 			remotealgos[count] = &algolist[i+1];
 			count++;
@@ -219,15 +221,17 @@ algo_type * buf_match_algo(buffer* buf, algo_type localalgos[]) {
 
 	/* iterate and find the first match */
 	for (i = 0; i < count; i++) {
+
 		len = strlen(remotealgos[i]);
+
 		for (j = 0; localalgos[j].name != NULL; j++) {
-			if (!localalgos[j].usable) {
-				continue;
-			}
-			if (len == strlen(localalgos[j].name) 
-					&& strcmp(localalgos[j].name, remotealgos[i]) == 0) {
-				ret = &localalgos[j];
-				goto out;
+			if (localalgos[j].usable) {
+				if (len == strlen(localalgos[j].name) 
+						&& strncmp(localalgos[j].name, 
+							remotealgos[i], len) == 0) {
+					ret = &localalgos[j];
+					goto out;
+				}
 			}
 		}
 	}
