@@ -26,11 +26,9 @@
 
 #define _ALGO_H_
 
-#include "options.h"
+#include "includes.h"
 #include "buffer.h"
 
-#include "libtomcrypt/mycrypt.h"
-
 struct Algo_Type {
 
 	unsigned char *name; /* identifying name */
@@ -65,7 +63,7 @@ struct dropbear_hash {
 };
 
 void crypto_init();
-int have_algo(char* algo, int algolen, algo_type algos[]);
+int have_algo(char* algo, size_t algolen, algo_type algos[]);
 algo_type * buf_match_algo(buffer* buf, algo_type localalgos[]);
 void buf_put_algolist(buffer * buf, algo_type localalgos[]);
 
