@@ -22,14 +22,11 @@
  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  * SOFTWARE. */
 
-#include <limits.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
+/* Process a pubkey auth request */
 
-#include "options.h"
+#include "includes.h"
 #include "session.h"
-#include "util.h"
+#include "dbutil.h"
 #include "buffer.h"
 #include "signkey.h"
 #include "auth.h"
@@ -42,8 +39,6 @@
 
 #define MIN_AUTHKEYS_LINE 10 /* "ssh-rsa AB" - short but doesn't matter */
 #define MAX_AUTHKEYS_LINE 1000 /* max length of a line in authkeys */
-#define MIN_AUTHKEYS_FILE MIN_AUTHKEYS_LINE
-#define MAX_AUTHKEYS_FILE 10000 /* should allow at least 20 authkeys */
 
 static int checkpubkey(unsigned char* algo, unsigned int algolen,
 		unsigned char* keyblob, unsigned int keybloblen);
@@ -53,7 +48,8 @@ static void send_msg_userauth_pk_ok(unsigned char* algo, unsigned int algolen,
 static int checkfileperm(char * filename);
 static int getauthline(buffer * line, FILE * authfile);
 
-/* process a pubkey auth request */
+/* process a pubkey auth request, sending success or failure message as
+ * appropriate */
 void pubkeyauth() {
 
 	unsigned char testkey; /* whether we're just checking if a key is usable */
@@ -133,12 +129,14 @@ out:
 	}
 	if (key) {
 		sign_key_free(key);
-
+		key = NULL;
 	}
 	TRACE(("leave pubkeyauth"));
 }
 
-/* Reply that the key is valid for auth */
+/* Reply that the key is valid for auth, this is sent when the user sends
+ * a straight copy of their pubkey to test, to avoid having to perform
+ * expensive signing operations with a worthless key */
 static void send_msg_userauth_pk_ok(unsigned char* algo, unsigned int algolen,
 		unsigned char* keyblob, unsigned int keybloblen) {
 
@@ -154,6 +152,8 @@ static void send_msg_userauth_pk_ok(unsigned char* algo, unsigned int algolen,
 
 }
 
+/* Checks whether a specified publickey (and associated algorithm) is an
+ * acceptable key for authentication */
 /* Returns DROPBEAR_SUCCESS if key is ok for auth, DROPBEAR_FAILURE otherwise */
 static int checkpubkey(unsigned char* algo, unsigned int algolen,
 		unsigned char* keyblob, unsigned int keybloblen) {
@@ -166,7 +166,7 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 	unsigned long decodekeylen;
 	unsigned char* filealgo = NULL;
 	unsigned int filealgolen;
-	int len;
+	unsigned int len, pos;
 	
 	TRACE(("enter checkpubkey"));
 
@@ -197,6 +197,7 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 	if (authfile == NULL) {
 		goto out;
 	}
+	TRACE(("checkpubkey: opened authorized_keys OK"));
 
 	line = buf_new(MAX_AUTHKEYS_LINE);
 
@@ -211,10 +212,12 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 
 		if (getauthline(line, authfile) == DROPBEAR_FAILURE) {
 			/* EOF reached */
+			TRACE(("checkpubkey: authorized_keys EOF reached"));
 			break;
 		}
 
 		if (line->len < MIN_AUTHKEYS_LINE) {
+			TRACE(("checkpubkey: line too short"));
 			continue; /* line is too short for it to be a valid key */
 		}
 
@@ -227,9 +230,20 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 		
 		/* check for space (' ') character */
 		if (buf_getbyte(line) != ' ') {
+			TRACE(("checkpubkey: space character expected, isn't there"));
 			continue;
 		}
 
+		/* truncate the line at the space after the base64 data */
+		pos = line->pos;
+		for (len = 0; line->pos < line->len; len++) {
+			if (buf_getbyte(line) == ' ') break;
+		}	
+		buf_setpos(line, pos);
+		buf_setlen(line, line->pos + len);
+
+		TRACE(("checkpubkey: line pos = %d len = %d", line->pos, line->len));
+
 		/* now we have the actual data */
 		decodekeylen = (line->len - line->pos) * 2;
 		decodekey = buf_new(decodekeylen);
@@ -237,14 +251,17 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 					line->len - line->pos,
 					buf_getwriteptr(decodekey, decodekey->size),
 					&decodekeylen) != CRYPT_OK) {
+			TRACE(("checkpubkey: base64 decode failed"));
 			continue;
 		}
+		TRACE(("checkpubkey: base64_decode success"));
 		buf_incrlen(decodekey, decodekeylen);
 		
 		/* compare the keys */
 		if (decodekeylen != keybloblen || memcmp(
 					buf_getptr(decodekey, decodekey->len),
 					keyblob, decodekey->len) != 0) {
+			TRACE(("checkpubkey: compare failed"));
 			continue;
 		}
 
@@ -252,6 +269,7 @@ static int checkpubkey(unsigned char* algo, unsigned int algolen,
 		 * itself match */
 		filealgo = buf_getstring(decodekey, &filealgolen);
 		if (filealgolen != algolen || memcmp(filealgo, algo, algolen) != 0) {
+			TRACE(("checkpubkey: algo match failed")); 
 			continue;
 		}
 
@@ -273,40 +291,47 @@ out:
 	}
 	m_free(filename);
 	m_free(filealgo);
-	TRACE(("leave checkpubkey"));
+	TRACE(("leave checkpubkey: ret=%d", ret));
 	return ret;
 }
 
 /* get a line from the file into buffer in the style expected for an
- * authkeys file, we stop after reaching a '=', but will read out to the
- * end of the file.
- * Will return DROPBEAR_SUCCESS if data is read, or DROPBEAR_FAILURE on EOF.
- * Note that it may return DROPBEAR_SUCCESS even with an empty line */
+ * authkeys file.
+ * Will return DROPBEAR_SUCCESS if data is read, or DROPBEAR_FAILURE on EOF.*/
 static int getauthline(buffer * line, FILE * authfile) {
 
-	int endofbase64 = 0;
-	int c;
-	int count = 0;
+	int c = EOF;
+
+	TRACE(("enter getauthline"));
+
 	buf_setpos(line, 0);
 	buf_setlen(line, 0);
-	for (;;) {
-		c = getc(authfile);
+
+	while (line->pos < line->size) {
+		c = fgetc(authfile); /*getc() is weird with some uClibc systems*/
 		if (c == EOF || c == '\n' || c == '\r') {
-			buf_setpos(line, 0);
-			if (count == 0 && c == EOF) {
-				return DROPBEAR_FAILURE;
-			} else {
-				return DROPBEAR_SUCCESS;
-			}
+			goto out;
 		}
-		if (c == '=') { /* base64 end char */
-			endofbase64 = 1;
-		}
-		if ((!endofbase64 || c == '=') && line->pos < line->size) {
-			buf_putbyte(line, (unsigned char)c);
-		} /* otherwise loop until EOL, ignoring extra */
-		count++;
+		buf_putbyte(line, (unsigned char)c);
 	}
+
+	TRACE(("leave getauthline: line too long"));
+	return DROPBEAR_FAILURE;
+
+out:
+
+	buf_setpos(line, 0);
+
+	/* if we didn't read anything before EOF or error, exit */
+	if (c == EOF && line->pos == 0) {
+		TRACE(("leave getauthline: failure"));
+		return DROPBEAR_FAILURE;
+	} else {
+		TRACE(("leave getauthline: success"));
+		return DROPBEAR_SUCCESS;
+	}
+
+	TRACE(("leave getauthline"));
 }	
 
 /* Returns DROPBEAR_SUCCESS if file permissions for pubkeys are ok,
@@ -337,19 +362,19 @@ static int checkpubkeyperms() {
 	strncpy(filename, ses.authstate.pw->pw_dir, len+1);
 
 	/* check ~ */
-	if (!checkfileperm(filename)) {
+	if (checkfileperm(filename) != DROPBEAR_SUCCESS) {
 		goto out;
 	}
 
 	/* check ~/.ssh */
 	strncat(filename, "/.ssh", 5); /* strlen("/.ssh") == 5 */
-	if (!checkfileperm(filename)) {
+	if (checkfileperm(filename) != DROPBEAR_SUCCESS) {
 		goto out;
 	}
 
 	/* now check ~/.ssh/authorized_keys */
 	strncat(filename, "/authorized_keys", 16);
-	if (!checkfileperm(filename)) {
+	if (checkfileperm(filename) != DROPBEAR_SUCCESS) {
 		goto out;
 	}
 
@@ -363,23 +388,31 @@ out:
 	return ret;
 }
 
-/* returns 1 on valid, 0 on invalid perms */
+/* Checks that a file is owned by the user or root, and isn't writable by
+ * group or other */
+/* returns DROPBEAR_SUCCESS or DROPBEAR_FAILURE */
 static int checkfileperm(char * filename) {
 	struct stat filestat;
 
+	TRACE(("enter checkfileperm(%s)", filename));
+
 	if (stat(filename, &filestat) != 0) {
-		return 0;
+		TRACE(("leave checkfileperm: stat() != 0"));
+		return DROPBEAR_FAILURE;
 	}
 	/* check ownership - user or root only*/
 	if (filestat.st_uid != ses.authstate.pw->pw_uid
 			&& filestat.st_uid != 0) {
-		return 0;
+		TRACE(("leave checkfileperm: wrong ownership"));
+		return DROPBEAR_FAILURE;
 	}
 	/* check permissions - don't want group or others +w */
 	if (filestat.st_mode & (S_IWGRP | S_IWOTH)) {
-		return 0;
+		TRACE(("leave checkfileperm: wrong perms"));
+		return DROPBEAR_FAILURE;
 	}
-	return 1;
+	TRACE(("leave checkfileperm: success"));
+	return DROPBEAR_SUCCESS;
 }
 
 
