@@ -22,18 +22,40 @@
  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  * SOFTWARE. */
 
-#include "options.h"
-#include "util.h"
-#include "libtomcrypt/mycrypt.h"
+/* Buffer handling routines, designed to avoid overflows/using invalid data */
+
+#include "includes.h"
+#include "dbutil.h"
 #include "buffer.h"
 
+/* Prevent integer overflows when incrementing buffer position/length.
+ * Calling functions should check arguments first, but this provides a
+ * backstop */
+#define BUF_MAX_INCR 1000000000
+#define BUF_MAX_SIZE 1000000000
+
+/* Create (malloc) a new buffer of size */
 buffer* buf_new(unsigned int size) {
 
-	buffer* ret;
+	buffer* buf;
 	
-	ret = (buffer*)m_malloc(sizeof(buffer));
-	buf_init(ret, size);
-	return ret;
+	if (size > BUF_MAX_SIZE) {
+		dropbear_exit("buf->size too big");
+	}
+
+	buf = (buffer*)m_malloc(sizeof(buffer));
+
+	if (size > 0) {
+		buf->data = (unsigned char*)m_malloc(size);
+	} else {
+		buf->data = NULL;
+	}
+
+	buf->size = size;
+	buf->pos = 0;
+	buf->len = 0;
+
+	return buf;
 
 }
 
@@ -51,24 +73,13 @@ void buf_burn(buffer* buf) {
 
 }
 
-/* initialise an already allocated buffer. The data won't be freed before
- * malloc */
-void buf_init(buffer* buf, unsigned int size) {
-
-	if (size > 0) {
-		buf->data = (unsigned char*)m_malloc(size);
-	} else {
-		buf->data = NULL;
-	}
-
-	buf->size = size;
-	buf->pos = 0;
-	buf->len = 0;
-}
-
 /* resize a buffer, pos and len will be repositioned if required */
 void buf_resize(buffer *buf, unsigned int newsize) {
 
+	if (newsize > BUF_MAX_SIZE) {
+		dropbear_exit("buf->size too big");
+	}
+
 	buf->data = m_realloc(buf->data, newsize);
 	buf->size = newsize;
 	buf->len = MIN(newsize, buf->len);
@@ -76,22 +87,19 @@ void buf_resize(buffer *buf, unsigned int newsize) {
 
 }
 
-/* create a copy of buf, allocating required memory etc */
-/* lenonly is a boolean flag specifying whether to set the size of the new
- * buffer to be just the len of the source buffer (1), or the size of the
- * source buffer (0) */
-buffer* buf_newcopy(buffer* buf, int lenonly) {
+/* Create a copy of buf, allocating required memory etc. */
+/* The new buffer is sized the same as the length of the source buffer. */
+buffer* buf_newcopy(buffer* buf) {
 	
 	buffer* ret;
-	unsigned int newsize;
 
-	newsize = lenonly ? buf->len : buf->size;
-	ret = buf_new(newsize);
+	ret = buf_new(buf->len);
 	ret->len = buf->len;
 	memcpy(ret->data, buf->data, buf->len);
 	return ret;
 }
 
+/* Set the length of the buffer */
 void buf_setlen(buffer* buf, unsigned int len) {
 	if (len > buf->size) {
 		dropbear_exit("bad buf_setlen");
@@ -99,12 +107,14 @@ void buf_setlen(buffer* buf, unsigned int len) {
 	buf->len = len;
 }
 
+/* Increment the length of the buffer */
 void buf_incrlen(buffer* buf, unsigned int incr) {
-	if (buf->len + incr > buf->size) {
+	if (incr > BUF_MAX_INCR || buf->len + incr > buf->size) {
 		dropbear_exit("bad buf_incrlen");
 	}
 	buf->len += incr;
 }
+/* Set the position of the buffer */
 void buf_setpos(buffer* buf, unsigned int pos) {
 
 	if (pos > buf->len) {
@@ -115,7 +125,7 @@ void buf_setpos(buffer* buf, unsigned int pos) {
 
 /* increment the postion by incr, increasing the buffer length if required */
 void buf_incrwritepos(buffer* buf, unsigned int incr) {
-	if (buf->pos + incr > buf->size) {
+	if (incr > BUF_MAX_INCR || buf->pos + incr > buf->size) {
 		dropbear_exit("bad buf_incrwritepos");
 	}
 	buf->pos += incr;
@@ -127,12 +137,15 @@ void buf_incrwritepos(buffer* buf, unsigned int incr) {
 /* increment the position by incr, negative values are allowed, to
  * decrement the pos*/
 void buf_incrpos(buffer* buf,  int incr) {
-	if (buf->pos + incr > buf->len || buf->pos + incr < 0) {
+	if (incr > BUF_MAX_INCR ||
+			(unsigned int)((int)buf->pos + incr) > buf->len 
+			|| ((int)buf->pos + incr) < 0) {
 		dropbear_exit("bad buf_incrpos");
 	}
 	buf->pos += incr;
 }
 
+/* Get a byte from the buffer and increment the pos */
 unsigned char buf_getbyte(buffer* buf) {
 
 	if (buf->pos >= buf->len) {
@@ -151,7 +164,8 @@ void buf_putbyte(buffer* buf, unsigned char val) {
 	buf->pos++;
 }
 
-/* returns an in-place pointer to the buffer, for boundschecking */
+/* returns an in-place pointer to the buffer, checking that
+ * the next len bytes from that position can be used */
 unsigned char* buf_getptr(buffer* buf, unsigned int len) {
 
 	if (buf->pos + len > buf->len) {
@@ -170,7 +184,7 @@ unsigned char* buf_getwriteptr(buffer* buf, unsigned int len) {
 	return &buf->data[buf->pos];
 }
 
-/* return a null-terminated string, it is malloced, so must be free()ed
+/* Return a null-terminated string, it is malloced, so must be free()ed
  * Note that the string isn't checked for null bytes, hence the retlen
  * may be longer than what is returned by strlen */
 unsigned char* buf_getstring(buffer* buf, unsigned int *retlen) {
@@ -193,6 +207,7 @@ unsigned char* buf_getstring(buffer* buf, unsigned int *retlen) {
 	return ret;
 }
 
+/* Get an uint32 from the buffer and increment the pos */
 unsigned int buf_getint(buffer* buf) {
 	unsigned int ret;
 
@@ -234,7 +249,7 @@ void buf_putmpint(buffer* buf, mp_int * mp) {
 
 	assert(mp != NULL);
 
-	if (SIGN(mp)) {
+	if (SIGN(mp) == MP_NEG) {
 		dropbear_exit("negative bignum");
 	}
 
