@@ -29,7 +29,7 @@
 #include "packet.h"
 #include "ssh.h"
 #include "buffer.h"
-#include "util.h"
+#include "dbutil.h"
 #include "channel.h"
 #include "chansession.h"
 #include "ssh.h"
@@ -49,12 +49,20 @@ static void send_msg_channel_data(struct Channel *channel, int isextended,
 		unsigned int exttype);
 static void send_msg_channel_eof(struct Channel *channel);
 static void send_msg_channel_close(struct Channel *channel);
-static void closechannel(struct Channel *channel);
+static void removechannel(struct Channel *channel);
 static void deletechannel(struct Channel *channel);
 static void send_exitsignalstatus(struct Channel *channel);
 static void checkinitdone(struct Channel *channel);
+static void checkclose(struct Channel *channel);
 
-/* initialise channels */
+static void closeinfd(struct Channel * channel);
+static void closeoutfd(struct Channel * channel, int fd);
+static void closechanfd(struct Channel *channel, int fd, int how);
+
+#define FD_UNINIT (-2)
+#define FD_CLOSED (-1)
+
+/* Initialise all the channels */
 void chaninitialise() {
 
 	/* may as well create space for a single channel */
@@ -74,14 +82,14 @@ void chancleanup() {
 	for (i = 0; i < ses.chansize; i++) {
 		if (ses.channels[i] != NULL) {
 			TRACE(("channel %d closing", i));
-			closechannel(ses.channels[i]);
+			removechannel(ses.channels[i]);
 		}
 	}
 	m_free(ses.channels);
 	TRACE(("leave chancleanup"));
 }
 
-/* create a new channel entry, send a reply confirm or failure */
+/* Create a new channel entry, send a reply confirm or failure */
 /* If remotechan, transwindow and transmaxpacket are not know (for a new
  * outgoing connection, with them to be filled on confirmation), they should
  * all be set to 0 */
@@ -123,17 +131,17 @@ struct Channel* newchannel(unsigned int remotechan, unsigned char type,
 	newchan = (struct Channel*)m_malloc(sizeof(struct Channel));
 	newchan->type = type;
 	newchan->index = i;
-	newchan->sentclosed = 0;
-	newchan->recveof = newchan->transeof = newchan->erreof = 0;
+	newchan->sentclosed = newchan->recvclosed = 0;
+	newchan->senteof = newchan->recveof = 0;
 
 	newchan->remotechan = remotechan;
 	newchan->transwindow = transwindow;
 	newchan->transmaxpacket = transmaxpacket;
 	
 	newchan->typedata = NULL;
-	newchan->infd = -1;
-	newchan->outfd = -1;
-	newchan->errfd = -1;
+	newchan->infd = FD_UNINIT;
+	newchan->outfd = FD_UNINIT;
+	newchan->errfd = FD_CLOSED; /* this isn't always set to start with */
 	newchan->initconn = 0;
 
 	newchan->writebuf = buf_new(RECV_MAXWINDOW);
@@ -161,26 +169,38 @@ void channelio(fd_set *readfd, fd_set *writefd) {
 	struct Channel *channel;
 	unsigned int i;
 
+	/* iterate through all the possible channels */
 	for (i = 0; i < ses.chansize; i++) {
 
 		channel = ses.channels[i];
-		if (channel == NULL || channel->sentclosed) {
+		if (channel == NULL) {
+			/* only process in-use channels */
 			continue;
 		}
 
-		/* read from program/pipe/etc stdout */
-		if (channel->outfd != -1 && channel->transeof == 0 &&
-				FD_ISSET(channel->outfd, readfd)) {
+		/* read from program/pipe stdout */
+		if (channel->outfd >= 0 && FD_ISSET(channel->outfd, readfd)) {
 			send_msg_channel_data(channel, 0, 0);
 		}
-		/* read from program/pipe stderr for interactive sessions */
-		if (channel->errfd != -1 && channel->erreof == 0 &&
-				FD_ISSET(channel->errfd, readfd)) {
+
+		/* read from program/pipe stderr */
+		if (channel->errfd >= 0 && FD_ISSET(channel->errfd, readfd)) {
 				send_msg_channel_data(channel, 1, SSH_EXTENDED_DATA_STDERR);
 		}
+
+		/* if we can read from the infd, it might be closed, so we try to
+		 * see if it has errors */
+		if (channel->infd >= 0 && channel->infd != channel->outfd
+				&& FD_ISSET(channel->infd, readfd)) {
+			int ret;
+			ret = write(channel->infd, NULL, 0);
+			if (ret < 0 && errno != EINTR && errno != EAGAIN) {
+				closeinfd(channel);
+			}
+		}
+
 		/* write to program/pipe stdin */
-		if (channel->infd != -1 && channel->recveof == 0 &&
-				FD_ISSET(channel->infd, writefd)) {
+		if (channel->infd >= 0 && FD_ISSET(channel->infd, writefd)) {
 			if (channel->initconn) {
 				checkinitdone(channel);
 				continue; /* Important not to use the channel after
@@ -189,25 +209,60 @@ void channelio(fd_set *readfd, fd_set *writefd) {
 				writechannel(channel);
 			}
 		}
-		
+	
 		/* handle any listening sockets */
 		if (channel->type == CHANNEL_ID_SESSION) {
 			struct ChanSess * chansess = (struct ChanSess*)channel->typedata;
 #ifndef DISABLE_X11FWD
-			if (chansess->x11fd != -1 && FD_ISSET(chansess->x11fd, readfd)) {
+			if (chansess->x11fd >= 0 && FD_ISSET(chansess->x11fd, readfd)) {
 				x11accept(chansess);
 			}
 #endif
 #ifndef DISABLE_AGENTFWD
-			if (chansess->agentfd != -1 && FD_ISSET(chansess->agentfd,readfd)) {
+			if (chansess->agentfd >= 0 && FD_ISSET(chansess->agentfd,readfd)) {
 				agentaccept(chansess);
 			}
 #endif
 		}
-	} /* foreach channel */
 
+		/* now handle any of the channel-closing type stuff */
+		checkclose(channel);
+
+	} /* foreach channel */
 }
 
+
+/* do all the EOF/close type stuff checking for a channel */
+static void checkclose(struct Channel *channel) {
+
+	if (!channel->sentclosed) {
+
+		/* check for exited */
+		if (channel->type == CHANNEL_ID_SESSION) {
+			if (((struct ChanSess*)channel->typedata)->exited) {
+				closeinfd(channel);
+			}
+		}
+
+		if (!channel->senteof
+			&& channel->outfd == FD_CLOSED 
+			&& channel->errfd == FD_CLOSED) {
+			send_msg_channel_eof(channel);
+		}
+
+		if (channel->infd == FD_CLOSED
+				&& channel->outfd == FD_CLOSED
+				&& channel->errfd == FD_CLOSED) {
+			send_msg_channel_close(channel);
+		}
+	}
+
+	if (channel->sentclosed && channel->recvclosed) {
+		removechannel(channel);
+	}
+}
+
+
 /* Check whether a deferred (EINPROGRESS) connect() was successful, and
  * if so, set up the channel properly. Otherwise, the channel is cleaned up, so
  * it is important that the channel reference isn't used after a call to this
@@ -263,6 +318,7 @@ static void send_msg_channel_close(struct Channel *channel) {
 
 	encrypt_packet();
 
+	channel->senteof = 1;
 	channel->sentclosed = 1;
 	TRACE(("leave send_msg_channel_close"));
 }
@@ -278,8 +334,7 @@ static void send_msg_channel_eof(struct Channel *channel) {
 
 	encrypt_packet();
 
-	/* we already know that trans/eof channels are closed */
-	send_msg_channel_close(channel);
+	channel->senteof = 1;
 
 	TRACE(("leave send_msg_channel_eof"));
 }
@@ -295,36 +350,32 @@ static void writechannel(struct Channel* channel) {
 
 	TRACE(("enter writechannel"));
 
-	if (channel->recveof || channel->sentclosed) {
-		TRACE(("leave writechannel: already recveof or sentclosed"));
-		return;
-	}
-
 	buf = channel->writebuf;
 	maxlen = buf->len - buf->pos;
 
 	len = write(channel->infd, buf_getptr(buf, maxlen), maxlen);
 	if (len <= 0) {
-		if (errno != EINTR) {
-
+		if (len < 0 && errno != EINTR) {
 			/* no more to write */
-			channel->recveof = 1;
-
-			/* if everything's closed then close it all up */
-			if (channel->transeof && 
-					(channel->erreof || channel->errfd == -1)) {
-				send_msg_channel_close(channel);
-			}
+			closeinfd(channel);
 		}
 		TRACE(("leave writechannel: len <= 0"));
 		return;
 	}
 	
-	/* extend the window */
-	/* TODO - this is inefficient */
 	if (len == maxlen) {
 		buf_setpos(buf, 0);
 		buf_setlen(buf, 0);
+
+		if (channel->recveof) {
+			/* we're closing up */
+			closeinfd(channel);
+			return;
+			TRACE(("leave writechannel: recveof set"));
+		}
+
+		/* extend the window if we're at the end*/
+		/* TODO - this is inefficient */
 		send_msg_channel_window_adjust(channel, buf->size
 				- channel->recvwindow);
 		channel->recvwindow = buf->size;
@@ -344,7 +395,7 @@ void setchannelfds(fd_set *readfd, fd_set *writefd) {
 	for (i = 0; i < ses.chansize; i++) {
 
 		channel = ses.channels[i];
-		if (channel == NULL || channel->sentclosed == 1) {
+		if (channel == NULL) {
 			continue;
 		}
 
@@ -352,34 +403,38 @@ void setchannelfds(fd_set *readfd, fd_set *writefd) {
 		if (channel->transwindow > 0) {
 
 			/* stdout */
-			if (channel->outfd != -1 && channel->transeof == 0) {
+			if (channel->outfd >= 0) {
 				/* there's space to read more from the program */
 				FD_SET(channel->outfd, readfd);
 			}
-			/* stderr for interactive sessions */
-			if (channel->errfd != -1 && channel->erreof == 0) {
+			/* stderr */
+			if (channel->errfd >= 0) {
 					FD_SET(channel->errfd, readfd);
 			}
 		}
+
+		if (channel->infd >= 0 && channel->infd != channel->outfd) {
+			FD_SET(channel->infd, readfd);
+		}
+
 		/* stdin */
-		if (channel->infd != -1 && channel->recveof == 0 &&
+		if (channel->infd >= 0 &&
 				(channel->writebuf->pos < channel->writebuf->len ||
 				 channel->initconn)) {
 			/* there's space to write more to the program */
 			FD_SET(channel->infd, writefd);
 		}
 
-		/* handle any listening sockets - should get optimised away if
-		 * we don't have x11 or agent fwd */
+		/* handle any listening sockets if we have x11 or agent fwd */
 		if (channel->type == CHANNEL_ID_SESSION) {
 			struct ChanSess * chansess = (struct ChanSess*)channel->typedata;
 #ifndef DISABLE_X11FWD
-			if (chansess->x11fd != -1) {
+			if (chansess->x11fd >= 0) {
 				FD_SET(chansess->x11fd, readfd);
 			}
 #endif
 #ifndef DISABLE_AGENTFWD
-			if (chansess->agentfd != -1) {
+			if (chansess->agentfd >= 0) {
 				FD_SET(chansess->agentfd, readfd);
 			}
 #endif
@@ -406,19 +461,8 @@ void recv_msg_channel_eof() {
 	}
 
 	channel->recveof = 1;
-
-	/* we should close the channel */
-	if (channel->type == CHANNEL_ID_X11 || channel->type == CHANNEL_ID_AGENT
-			|| channel->type == CHANNEL_ID_TCPDIRECT) {
-		shutdown(channel->infd, 0);
-	} else {
-		close(channel->infd);
-	}
-	channel->infd = -1;
-
-	if (channel->transeof && (channel->erreof || channel->errfd == -1)
-			&& !channel->sentclosed) {
-		send_msg_channel_close(channel);
+	if (channel->writebuf->len == 0) {
+		closeinfd(channel);
 	}
 
 	TRACE(("leave recv_msg_channel_eof"));
@@ -442,22 +486,23 @@ void recv_msg_channel_close() {
 		dropbear_exit("Close for unknown channel");
 	}
 
-	if (!channel->sentclosed) {
-		send_msg_channel_close(channel);
+	channel->recveof = 1;
+	channel->recvclosed = 1;
+
+	if (channel->sentclosed) {
+		removechannel(channel);
 	}
 
-	closechannel(channel);
-	
 	TRACE(("leave recv_msg_channel_close"));
 }
 
 /* Remove a channel entry, this is only executed after both sides have sent
  * channel close */
-static void closechannel(struct Channel * channel) {
+static void removechannel(struct Channel * channel) {
 
-	TRACE(("enter closechannel"));
+	TRACE(("enter removechannel"));
 	TRACE(("channel index is %d", channel->index));
-	
+
 	buf_free(channel->writebuf);
 
 	/* close the FDs in case they haven't been done
@@ -467,14 +512,15 @@ static void closechannel(struct Channel * channel) {
 
 	if (channel->type == CHANNEL_ID_SESSION) {
 		closechansess(channel);
+		close(channel->errfd);
 	}
 
 	deletechannel(channel);
 
-	TRACE(("leave closechannel"));
+	TRACE(("leave removechannel"));
 }
 
-/* Just remove a channel entry, don't try to close descriptors etc */
+/* Remove a channel entry */
 static void deletechannel(struct Channel *channel) {
 
 	ses.channels[channel->index] = NULL;
@@ -532,25 +578,16 @@ static void send_msg_channel_data(struct Channel *channel, int isextended,
 	unsigned int maxlen;
 	int fd;
 
-	TRACE(("enter send_msg_channel_data"));
-	TRACE(("extended = %d type = %d", isextended, exttype));
+/*	TRACE(("enter send_msg_channel_data"));
+	TRACE(("extended = %d type = %d", isextended, exttype));*/
 
 	CHECKCLEARTOWRITE();
 
 	assert(!channel->sentclosed);
 
 	if (isextended) {
-		if (channel->erreof) {
-			TRACE(("leave send_msg_channel_data: erreof already set"));
-			return;
-		}
-		assert(exttype == SSH_EXTENDED_DATA_STDERR);
 		fd = channel->errfd;
 	} else {
-		if (channel->transeof) {
-			TRACE(("leave send_msg_channel_data: transeof already set"));
-			return;
-		}
 		fd = channel->outfd;
 	}
 	assert(fd >= 0);
@@ -558,8 +595,8 @@ static void send_msg_channel_data(struct Channel *channel, int isextended,
 	maxlen = MIN(channel->transwindow, channel->transmaxpacket);
 	/* -(1+4+4) is SSH_MSG_CHANNEL_DATA, channel number, string length, and 
 	 * exttype if is extended */
-	maxlen = MIN(maxlen, ses.writepayload->size 
-			- 1 - 4 - 4 - (isextended ? 4 : 0));
+	maxlen = MIN(maxlen, 
+			ses.writepayload->size - 1 - 4 - 4 - (isextended ? 4 : 0));
 	if (maxlen == 0) {
 		TRACE(("leave send_msg_channel_data: no window"));
 		return; /* the data will get written later */
@@ -569,23 +606,12 @@ static void send_msg_channel_data(struct Channel *channel, int isextended,
 	buf = buf_new(maxlen);
 	len = read(fd, buf_getwriteptr(buf, maxlen), maxlen);
 	if (len <= 0) {
-		/* on error etc, send eof */
-		if (errno != EINTR) {
-			
-			if (isextended) {
-				channel->erreof = 1;
-			} else {
-				channel->transeof = 1;
-			}
-			
-			if ((channel->erreof || channel->errfd == -1)
-					&& channel->transeof) {
-				send_msg_channel_eof(channel);
-			}
+		/* on error/eof, send eof */
+		if (len == 0 || errno != EINTR) {
+			closeoutfd(channel, fd);
+			TRACE(("leave send_msg_channel_data: read err"));
 		}
 		buf_free(buf);
-		TRACE(("leave send_msg_channel_data: len <= 0, erreof %d transeof %d",
-					channel->erreof, channel->transeof));
 		return;
 	}
 	buf_incrlen(buf, len);
@@ -623,11 +649,16 @@ void recv_msg_channel_data() {
 	chan = buf_getint(ses.payload);
 	channel = getchannel(chan);
 	if (channel == NULL) {
-		/* disconnect ? */
 		dropbear_exit("Unknown channel");
 	}
 
-	assert(channel->infd != -1);
+	if (channel->recveof) {
+		dropbear_exit("received data after eof");
+	}
+
+ 	if (channel->infd < 0) {
+		dropbear_exit("received data with bad infd");
+	}
 
 	datalen = buf_getint(ses.payload);
 
@@ -648,6 +679,8 @@ void recv_msg_channel_data() {
 	buf_setpos(channel->writebuf, pos); /* revert pos */
 
 	channel->recvwindow -= datalen;
+
+	/* matt - this might be for later */
 /*	if (channel->recvwindow < RECV_MINWINDOW) {
 		send_msg_channel_window_adjust(channel, 
 				RECV_MAXWINDOW - channel->recvwindow);
@@ -670,7 +703,7 @@ void recv_msg_channel_window_adjust() {
 	channel = getchannel(chan);
 
 	if (channel == NULL) {
-		dropbear_exit("Unknown channel"); /* TODO - disconnect */
+		dropbear_exit("Unknown channel");
 	}
 	
 	incr = buf_getint(ses.payload);
@@ -899,6 +932,54 @@ void recv_msg_channel_open_failure() {
 		dropbear_exit("Unknown channel");
 	}
 
-	closechannel(channel);
+	removechannel(channel);
 }
+
+/* close a stdout/stderr fd */
+static void closeoutfd(struct Channel * channel, int fd) {
+
+	/* don't close it if it is the same as infd,
+	 * unless infd is already set -1 */
+	TRACE(("enter closeoutfd"));
+	closechanfd(channel, fd, 0);
+	TRACE(("leave closeoutfd"));
+}
+
+/* close a stdin fd */
+static void closeinfd(struct Channel * channel) {
+
+	TRACE(("enter closeinfd"));
+	closechanfd(channel, channel->infd, 1);
+	TRACE(("leave closeinfd"));
+}
+
+/* close a fd, how is 0 for stdout/stderr, 1 for stdin */
+static void closechanfd(struct Channel *channel, int fd, int how) {
+
+	int closein = 0, closeout = 0;
+
+	if (channel->type == CHANNEL_ID_X11 || channel->type == CHANNEL_ID_AGENT
+			|| channel->type == CHANNEL_ID_TCPDIRECT) {
+		shutdown(fd, how);
+		if (how == 0) {
+			closeout = 1;
+		} else {
+			closein = 1;
+		}
+	} else {
+		close(fd);
+		closein = closeout = 1;
+	}
+
+	if (closeout && fd == channel->errfd) {
+		channel->errfd = FD_CLOSED;
+	}
+	if (closeout && fd == channel->outfd) {
+		channel->outfd = FD_CLOSED;
+	}
+	if (closein && fd == channel->infd) {
+		channel->infd = FD_CLOSED;
+	}
+}
+
 #endif /* USE_LISTENERS */
