@@ -43,8 +43,6 @@ void printhex(unsigned char* buf, int len);
 char * stripcontrol(const char * text);
 unsigned char * getaddrstring(struct sockaddr * addr);
 char* getaddrhostname(struct sockaddr * addr);
-int readln(int fd, char* buf, int count);
-int writeln(int fd, const char* str);
 int buf_readfile(buffer* buf, const char* filename);
 
 int m_close(int fd);
