@@ -26,7 +26,7 @@
 #include "runopts.h"
 #include "signkey.h"
 #include "buffer.h"
-#include "util.h"
+#include "dbutil.h"
 #include "algo.h"
 
 static sign_key * loadhostkeys(const char * dsskeyfile,
@@ -42,26 +42,47 @@ static void printhelp(const char * progname) {
 					"-b bannerfile     Display the contents of bannerfile"
 					" before user login\n"
 					"                  (default: none)\n"
+#ifdef DROPBEAR_DSS
 					"-d dsskeyfile     Use dsskeyfile for the dss host key\n"
 					"                  (default: %s)\n"
+#endif
+#ifdef DROPBEAR_RSA
 					"-r rsakeyfile     Use rsakeyfile for the rsa host key\n"
 					"                  (default: %s)\n"
+#endif
 					"-F                Don't fork into background\n"
-					"-p port           Listen on specified tcp port\n"
-					"                  (default %d)\n",
-					DROPBEAR_VERSION,
-					progname, DSS_PRIV_FILENAME, RSA_PRIV_FILENAME,
-					DROPBEAR_PORT);
+#ifdef DISABLE_SYSLOG
+					"(Syslog support not compiled in, using stderr)\n"
+#else
+					"-E                Log to stderr rather than syslog\n"
+#endif
+#ifdef DO_MOTD
+					"-m                Don't display the motd on login\n"
+#endif
+					"-w                Disallow root logins\n"
+					"-p port           Listen on specified tcp port, up to %d can be specified\n"
+					"                  (default %d if none specified)\n"
+/*					"-4/-6             Disable listening on ipv4/ipv6 respectively\n"*/
+
+					,DROPBEAR_VERSION, progname,
+#ifdef DROPBEAR_DSS
+					DSS_PRIV_FILENAME,
+#endif
+#ifdef DROPBEAR_RSA
+					RSA_PRIV_FILENAME,
+#endif
+					DROPBEAR_MAX_PORTS, DROPBEAR_PORT);
 }
 
 /* returns NULL on failure, or a pointer to a freshly allocated
  * runopts structure */
 runopts * getrunopts(int argc, char ** argv) {
 
-	int i;
+	unsigned int i;
 	char ** next = 0;
 	runopts * opts;
-	char * portstring = NULL;
+	unsigned int portnum = 0;
+	char *portstring[DROPBEAR_MAX_PORTS];
 	unsigned int longport;
 
 	/* see printhelp() for options */
@@ -71,6 +92,17 @@ runopts * getrunopts(int argc, char ** argv) {
 	opts->bannerfile = NULL;
 	opts->banner = NULL;
 	opts->forkbg = 1;
+	opts->norootlogin = 0;
+	/* not yet
+	opts->ipv4 = 1;
+	opts->ipv6 = 1;
+	*/
+#ifdef DO_MOTD
+	opts->domotd = 1;
+#endif
+#ifndef DISABLE_SYSLOG
+	usingsyslog = 1;
+#endif
 
 	for (i = 1; i < argc; i++) {
 		if (next) {
@@ -87,22 +119,52 @@ runopts * getrunopts(int argc, char ** argv) {
 				case 'b':
 					next = &opts->bannerfile;
 					break;
+#ifdef DROPBEAR_DSS
 				case 'd':
 					next = &opts->dsskeyfile;
 					break;
+#endif
+#ifdef DROPBEAR_RSA
 				case 'r':
 					next = &opts->rsakeyfile;
 					break;
+#endif
 				case 'F':
 					opts->forkbg = 0;
 					break;
+#ifndef DISABLE_SYSLOG
+				case 'E':
+					usingsyslog = 0;
+					break;
+#endif
 				case 'p':
-					next = &portstring;
+					if (portnum < DROPBEAR_MAX_PORTS) {
+						portstring[portnum] = NULL;
+						next = &portstring[portnum];
+						portnum++;
+					}
+					break;
+#ifdef DO_MOTD
+				/* motd is displayed by default, -m turns it off */
+				case 'm':
+					opts->domotd = 0;
+					break;
+#endif
+				case 'w':
+					opts->norootlogin = 1;
 					break;
 				case 'h':
 					printhelp(argv[0]);
 					exit(EXIT_FAILURE);
 					break;
+					/*
+				case '4':
+					opts->ipv4 = 0;
+					break;
+				case '6':
+					opts->ipv6 = 0;
+					break;
+					*/
 				default:
 					fprintf(stderr, "Unknown argument %s\n", argv[i]);
 					printhelp(argv[0]);
@@ -140,19 +202,53 @@ runopts * getrunopts(int argc, char ** argv) {
 		buf_setpos(opts->banner, 0);
 	}
 
-	if (portstring) {
-		longport = atoi(portstring);
-		if (longport > 65534 || longport < 1) {
-			dropbear_exit("Bad port %s", portstring);
-		}
-		opts->port = (uint16_t)longport;
+	/* not yet
+	if (!(opts->ipv4 || opts->ipv6)) {
+		fprintf(stderr, "You can't disable ipv4 and ipv6.\n");
+		exit(1);
+	}
+	*/
+
+	/* create the array of listening ports */
+	if (portnum == 0) {
+		/* non specified */
+		opts->portcount = 1;
+		opts->ports = m_malloc(sizeof(uint16_t));
+		opts->ports[0] = DROPBEAR_PORT;
 	} else {
-		opts->port = DROPBEAR_PORT;
+		opts->portcount = portnum;
+		opts->ports = (uint16_t*)m_malloc(sizeof(uint16_t)*portnum);
+		for (i = 0; i < portnum; i++) {
+			if (portstring[i]) {
+				longport = atoi(portstring[i]);
+					if (longport <= 65535 && longport > 0) {
+						opts->ports[i] = (uint16_t)longport;
+						continue;
+					}
+			}
+			fprintf(stderr, "Bad port '%s'\n",
+					portstring[i] ? portstring[i] : "null");
+		}
 	}
 
 	return opts;
 }
 
+void freerunopts(runopts* opts) {
+
+	if (!opts) {
+		return;
+	}
+
+	if (opts->hostkey) {
+		sign_key_free(opts->hostkey);
+		opts->hostkey = NULL;
+	}
+
+	m_free(opts->ports);
+	m_free(opts);
+}
+
 /* returns success or failure */
 static int readhostkey(const char * filename, sign_key * hostkey, int type) {
 
@@ -210,9 +306,9 @@ static sign_key * loadhostkeys(const char * dsskeyfile,
 		&& hostkey->dsskey == NULL
 #endif
 #ifdef DROPBEAR_RSA
-		&& hostkey->rsakey == NULL)
+		&& hostkey->rsakey == NULL
 #endif
-		{
+		) {
 		dropbear_exit("No hostkeys available");
 	}
 
