@@ -25,17 +25,17 @@
 #include "includes.h"
 
 #ifndef DISABLE_X11FWD
-
 #include "x11fwd.h"
 #include "session.h"
 #include "ssh.h"
-#include "util.h"
+#include "dbutil.h"
 #include "chansession.h"
 #include "channel.h"
 #include "packet.h"
 #include "buffer.h"
 
-#define X11BASEPORT 6010
+#define X11BASEPORT 6000
+#define X11BINDBASE 6010
 
 static int bindport(int fd);
 static int send_msg_channel_open_x11(int fd, struct sockaddr_in* addr);
@@ -119,15 +119,16 @@ void x11setauth(struct ChanSess *chansess) {
 
 	char display[20]; /* space for "localhost:12345.123" */
 	FILE * authprog;
+	int val;
 
 	if (chansess->x11fd == -1) {
 		return;
 	}
 
 	/* create the DISPLAY string */
-	if (snprintf(display, sizeof(display), "localhost:%d.%d",
-			chansess->x11port - 6000, chansess->x11screennum) 
-			>= sizeof(display)) {
+	val = snprintf(display, sizeof(display), "localhost:%d.%d",
+			chansess->x11port - X11BASEPORT, chansess->x11screennum);
+	if (val < 0 || val >= (int)sizeof(display)) {
 		/* string was truncated */
 		return;
 	}
@@ -135,9 +136,9 @@ void x11setauth(struct ChanSess *chansess) {
 	addnewvar("DISPLAY", display);
 
 	/* create the xauth string */
-	if (snprintf(display, sizeof(display), "unix:%d.%d",
-			chansess->x11port - 6000, chansess->x11screennum) 
-			>= sizeof(display)) {
+	val = snprintf(display, sizeof(display), "unix:%d.%d",
+			chansess->x11port - X11BASEPORT, chansess->x11screennum);
+	if (val < 0 || val >= (int)sizeof(display)) {
 		/* string was truncated */
 		return;
 	}
@@ -151,7 +152,6 @@ void x11setauth(struct ChanSess *chansess) {
 	} else {
 		fprintf(stderr, "Failed to run %s\n", XAUTH_COMMAND);
 	}
-
 }
 
 void x11cleanup(struct ChanSess * chansess) {
@@ -184,17 +184,18 @@ static int send_msg_channel_open_x11(int fd, struct sockaddr_in* addr) {
 }
 
 /* returns the port bound to, or -1 on failure.
- * Will attempt to bind to a port X11BASEPORT (6010 usually) or upwards */
+ * Will attempt to bind to a port X11BINDBASE (6010 usually) or upwards */
 static int bindport(int fd) {
 
 	struct sockaddr_in addr;
 	uint16_t port;
 
+	memset((void*)&addr, 0x0, sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
 
 	/* if we can't find one in 2000 ports free, something's wrong */
-	for (port = X11BASEPORT; port < X11BASEPORT + 2000; port++) {
+	for (port = X11BINDBASE; port < X11BINDBASE + 2000; port++) {
 		addr.sin_port = htons(port);
 		if (bind(fd, (struct sockaddr*)&addr, 
 					sizeof(struct sockaddr_in)) == 0) {
@@ -206,6 +207,7 @@ static int bindport(int fd) {
 			continue;
 		}
 		/* otherwise it was an error we don't know about */
+		dropbear_log(LOG_DEBUG, "failed to bind x11 socket");
 		break;
 	}
 	return -1;
